/*
 *    Copyright 2018 Niek Haarman
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package com.nhaarman.acorn.navigation.experimental

import com.nhaarman.acorn.navigation.DisposableHandle
import com.nhaarman.acorn.navigation.Navigator
import com.nhaarman.acorn.navigation.SavableNavigator
import com.nhaarman.acorn.navigation.SingleSceneNavigator
import com.nhaarman.acorn.navigation.StackNavigator
import com.nhaarman.acorn.navigation.TransitionData
import com.nhaarman.acorn.navigation.experimental.CompositeParallelNavigatorTest.TestDestination.Destination1
import com.nhaarman.acorn.navigation.experimental.CompositeParallelNavigatorTest.TestDestination.Destination2
import com.nhaarman.acorn.navigation.experimental.CompositeParallelNavigatorTest.TestDestination.Destination3
import com.nhaarman.acorn.navigation.experimental.CompositeParallelNavigatorTest.TestDestination.values
import com.nhaarman.acorn.presentation.Container
import com.nhaarman.acorn.presentation.Scene
import com.nhaarman.acorn.state.NavigatorState
import com.nhaarman.acorn.state.SavedState
import com.nhaarman.acorn.state.SceneState
import com.nhaarman.acorn.state.navigatorState
import com.nhaarman.expect.expect
import com.nhaarman.mockitokotlin2.any
import com.nhaarman.mockitokotlin2.anyOrNull
import com.nhaarman.mockitokotlin2.argumentCaptor
import com.nhaarman.mockitokotlin2.eq
import com.nhaarman.mockitokotlin2.inOrder
import com.nhaarman.mockitokotlin2.mock
import com.nhaarman.mockitokotlin2.never
import com.nhaarman.mockitokotlin2.spy
import com.nhaarman.mockitokotlin2.times
import com.nhaarman.mockitokotlin2.verify
import com.nhaarman.mockitokotlin2.verifyNoMoreInteractions
import kotlin.reflect.KClass
import org.junit.jupiter.api.BeforeEach
import org.junit.jupiter.api.Nested
import org.junit.jupiter.api.Test

@ExperimentalCompositeParallelNavigator
internal class CompositeParallelNavigatorTest {

    private val navigator1Scene1 = spy(SavableTestScene(11))
    private val navigator1Scene2 = spy(SavableTestScene(12))
    private val navigator2Scene1 = spy(SavableTestScene(21))
    private val navigator2Scene2 = spy(SavableTestScene(22))
    private val navigator3Scene1 = spy(SavableTestScene(31))

    private val extraNavigator = spy(SavableTestScene(99))

    private var navigator1 = spy(SavableTestStackNavigator(listOf(navigator1Scene1)))
    private var navigator2 = spy(SavableTestStackNavigator(listOf(navigator2Scene1)))
    private var navigator3 = spy(SavableTestStackNavigator(listOf(navigator3Scene1)))

    private val navigator = TestCompositeParallelNavigator()
    private val listener = mock<Navigator.Events>()

    @Nested
    inner class NavigatorStates {

        @Nested
        inner class InactiveNavigator {

            @Test
            fun `navigator is not finished`() {
                /* When */
                navigator.addNavigatorEventsListener(listener)

                /* Then */
                verify(listener, never()).finished()
            }

            @Test
            fun `navigator is not destroyed`() {
                /* Then */
                expect(navigator.isDestroyed()).toBe(false)
            }

            @Test
            fun `added listener does not get notified of scene`() {
                /* When */
                navigator.addNavigatorEventsListener(listener)

                /* Then */
                verify(listener, never()).scene(any(), any())
            }

            @Test
            fun `selecting a destination does not notify listeners of scene`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)

                /* When */
                navigator.select(Destination2)

                /* Then */
                verify(listener, never()).scene(any(), any())
            }

            @Test
            fun `onBackPressed for a single scene notifies listeners of finished`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)

                /* When */
                val result = navigator.onBackPressed()

                /* Then */
                expect(result).toBe(true)
                verify(listener).finished()
            }

            @Test
            fun `onBackPressed for a single scene does not notify scene`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)

                /* When */
                val result = navigator.onBackPressed()

                /* Then */
                expect(result).toBe(true)
                verify(listener, never()).scene(any(), any())
            }

            @Test
            fun `onBackPressed after navigator is destroyed does not notify listeners`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.onStart()
                navigator.onDestroy()

                /* When */
                val result = navigator.onBackPressed()

                /* Then */
                expect(result).toBe(false)
                verify(listener, never()).finished()
            }

            @Test
            fun `onBackPressed for two scenes in first destination does notify scenes`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator1.push(navigator1Scene2)

                /* When */
                val result = navigator.onBackPressed()

                /* Then */
                expect(result).toBe(true)
                listener.inOrder {
                    verifyNoMoreInteractions()
                }
            }
        }

        @Nested
        inner class ActiveNavigator {

            @Test
            fun `navigator is not finished`() {
                /* Given */
                navigator.onStart()

                /* When */
                navigator.addNavigatorEventsListener(listener)

                /* Then */
                verify(listener, never()).finished()
            }

            @Test
            fun `navigator is not destroyed`() {
                /* Given */
                navigator.onStart()

                /* Then */
                expect(navigator.isDestroyed()).toBe(false)
            }

            @Test
            fun `added listener gets notified of initial scene`() {
                /* Given */
                navigator.onStart()

                /* When */
                navigator.addNavigatorEventsListener(listener)

                /* Then */
                verify(listener).scene(navigator1Scene1, null)
            }

            @Test
            fun `removed listener does not get notified of scene`() {
                /* Given */
                val disposable = navigator.addNavigatorEventsListener(listener)
                disposable.dispose()

                /* When */
                navigator.onStart()

                /* Then */
                verify(listener, never()).scene(any(), anyOrNull())
            }

            @Test
            fun `non disposed listener is not disposed`() {
                /* Given */
                val disposable = navigator.addNavigatorEventsListener(listener)

                /* Then */
                expect(disposable.isDisposed()).toBe(false)
            }

            @Test
            fun `disposed listener is disposed`() {
                /* Given */
                val disposable = navigator.addNavigatorEventsListener(listener)

                /* When */
                disposable.dispose()

                /* Then */
                expect(disposable.isDisposed()).toBe(true)
            }

            @Test
            fun `starting navigator notifies listeners of initial scene`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)

                /* When */
                navigator.onStart()

                /* Then */
                verify(listener).scene(navigator1Scene1, null)
            }

            @Test
            fun `starting navigator multiple times notifies listeners of scene only once`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)

                /* When */
                navigator.onStart()
                navigator.onStart()

                /* Then */
                verify(listener, times(1)).scene(any(), anyOrNull())
            }

            @Test
            fun `starting navigator second time in a callback only notifies listener once`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)

                /* When */
                navigator.addNavigatorEventsListener(object : Navigator.Events {
                    override fun scene(scene: Scene<out Container>, data: TransitionData?) {
                        navigator.onStart()
                    }

                    override fun finished() {
                    }
                })
                navigator.onStart()

                /* Then */
                verify(listener, times(1)).scene(any(), anyOrNull())
            }

            @Test
            fun `starting navigator does not finish`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)

                /* When */
                navigator.onStart()

                /* Then */
                verify(listener, never()).finished()
            }

            @Test
            fun `changing destination notifies listeners of scene`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.onStart()

                /* When */
                navigator.select(Destination2)

                /* Then */
                verify(listener).scene(eq(navigator2Scene1), anyOrNull())
            }

            @Test
            fun `changing to the same destination twice notifies listeners of scene only once`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.onStart()

                /* When */
                navigator.select(Destination2)
                navigator.select(Destination2)

                /* Then */
                verify(listener, times(1)).scene(eq(navigator2Scene1), anyOrNull())
            }

            @Test
            fun `start navigator after destination changed notifies pushed scene`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.select(Destination2)

                /* When */
                navigator.onStart()

                /* Then */
                verify(listener).scene(navigator2Scene1)
            }

            @Test
            fun `onBackPressed for a single scene in first destination notifies finished`() {
                /* Given */
                navigator.onStart()
                navigator.addNavigatorEventsListener(listener)

                /* When */
                val result = navigator.onBackPressed()

                /* Then */
                expect(result).toBe(true)
                verify(listener).finished()
            }

            @Test
            fun `onBackPressed for a single scene in second destination does not notify finished`() {
                /* Given */
                navigator.onStart()
                navigator.addNavigatorEventsListener(listener)
                navigator.select(Destination2)

                /* When */
                val result = navigator.onBackPressed()

                /* Then */
                expect(result).toBe(true)
                verify(listener, never()).finished()
            }

            @Test
            fun `onBackPressed for a single scene in first destination does not notify scene`() {
                /* Given */
                navigator.onStart()
                navigator.addNavigatorEventsListener(listener)

                /* When */
                val result = navigator.onBackPressed()

                /* Then */
                expect(result).toBe(true)
                listener.inOrder {
                    verify().scene(navigator1Scene1)
                    verify().finished()
                    verifyNoMoreInteractions()
                }
            }

            @Test
            fun `onBackPressed for a single scene in second destination does notify scene`() {
                /* Given */
                navigator.onStart()
                navigator.addNavigatorEventsListener(listener)
                navigator.select(Destination2)

                /* When */
                val result = navigator.onBackPressed()

                /* Then */
                expect(result).toBe(true)
                listener.inOrder {
                    verify().scene(navigator1Scene1)
                    verify().scene(navigator2Scene1)
                    verify().scene(navigator1Scene1)
                    verifyNoMoreInteractions()
                }
            }

            @Test
            fun `onBackPressed for two scenes in first destination does notify scene`() {
                /* Given */
                navigator.onStart()
                navigator.addNavigatorEventsListener(listener)
                navigator1.push(navigator1Scene2)

                /* When */
                val result = navigator.onBackPressed()

                /* Then */
                expect(result).toBe(true)
                listener.inOrder {
                    verify().scene(navigator1Scene1)
                    verify().scene(navigator1Scene2, TransitionData.forwards)
                    verify().scene(navigator1Scene1, TransitionData.backwards)
                    verifyNoMoreInteractions()
                }
            }
        }

        @Nested
        inner class StoppedNavigator {

            @Test
            fun `stopping navigator does not finish`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)

                /* When */
                navigator.onStop()

                /* Then */
                verify(listener, never()).finished()
            }

            @Test
            fun `stopped navigator is not destroyed`() {
                /* When */
                navigator.onStop()

                /* Then */
                expect(navigator.isDestroyed()).toBe(false)
            }
        }

        @Nested
        inner class DestroyedNavigator {

            @Test
            fun `destroying navigator does not finish`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)

                /* When */
                navigator.onDestroy()

                /* Then */
                verify(listener, never()).finished()
            }

            @Test
            fun `isDestroyed returns true`() {
                /* When */
                navigator.onDestroy()

                /* Then */
                expect(navigator.isDestroyed()).toBe(true)
            }

            @Test
            fun `changing destination for destroyed navigator does not notify listeners`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.onDestroy()

                /* When */
                navigator.select(Destination2)

                /* Then */
                verify(listener, never()).scene(any(), any())
            }
        }
    }

    @Nested
    inner class SingleChildNavigatorInteraction {

        @Test
        fun `starting navigator starts initial navigator`() {
            /* When */
            navigator.onStart()

            /* Then */
            navigator1.inOrder {
                verify().onStart()
                verifyNoMoreInteractions()
            }
        }

        @Test
        fun `starting navigator multiple times starts initial navigator only once`() {
            /* When */
            navigator.onStart()
            navigator.onStart()

            /* Then */
            verify(navigator1, times(1)).onStart()
        }

        @Test
        fun `stopping an inactive navigator does not stop child navigator`() {
            /* When */
            navigator.onStop()

            /* Then */
            verify(navigator1, never()).onStop()
        }

        @Test
        fun `stopping an active navigator stops child navigator`() {
            /* Given */
            navigator.onStart()

            /* When */
            navigator.onStop()

            /* Then */
            navigator1.inOrder {
                verify().onStart()
                verify().onStop()
            }
        }

        @Test
        fun `destroy an inactive navigator does not stop child navigator`() {
            /* When */
            navigator.onDestroy()

            /* Then */
            verify(navigator1, never()).onStop()
        }

        @Test
        fun `destroy an inactive navigator does destroy child navigator`() {
            /* When */
            navigator.onDestroy()

            /* Then */
            verify(navigator1).onDestroy()
        }

        @Test
        fun `destroy an active navigator stops and destroys child navigator`() {
            /* Given */
            navigator.onStart()

            /* When */
            navigator.onDestroy()

            /* Then */
            navigator1.inOrder {
                verify().onStart()
                verify().onStop()
                verify().onDestroy()
            }
        }

        @Test
        fun `starting a destroyed navigator does not start child navigator`() {
            /* Given */
            navigator.onDestroy()

            /* When */
            navigator.onStart()

            /* Then */
            verify(navigator1).onDestroy()
            verify(navigator1, never()).onStart()
        }

        @Test
        fun `stopping a destroyed navigator does not start child navigator`() {
            /* Given */
            navigator.onDestroy()

            /* When */
            navigator.onStop()

            /* Then */
            verify(navigator1).onDestroy()
            verify(navigator1, never()).onStop()
        }

        @Test
        fun `destroying a destroyed navigator only destroys child navigator once`() {
            /* Given */
            navigator.onDestroy()

            /* When */
            navigator.onDestroy()

            /* Then */
            verify(navigator1, times(1)).onDestroy()
        }

        @Test
        fun `starting an inactive navigator results in scene emission`() {
            /* Given */
            navigator.addNavigatorEventsListener(listener)

            /* When */
            navigator.onStart()

            /* Then */
            verify(listener).scene(eq(navigator1Scene1), anyOrNull())
            verifyNoMoreInteractions(listener)
        }

        @Test
        fun `an inactive navigator does not propagate child navigator emitting scene`() {
            /* Given */
            navigator.addNavigatorEventsListener(listener)

            /* When */
            navigator1.push(navigator1Scene2)

            /* Then */
            verify(listener, never()).scene(any(), anyOrNull())
        }

        @Test
        fun `an active navigator propagates child navigator emitting scene`() {
            /* Given */
            navigator.addNavigatorEventsListener(listener)
            navigator.onStart()

            /* When */
            navigator1.push(navigator1Scene2)

            /* Then */
            listener.inOrder {
                verify().scene(eq(navigator1Scene1), anyOrNull())
                verify().scene(eq(navigator1Scene2), anyOrNull())
            }
        }

        @Test
        fun `a destroyed navigator does not propagate child navigator emitting scene`() {
            /* Given */
            navigator.addNavigatorEventsListener(listener)
            navigator.onStart()
            navigator.onStop()
            navigator.onDestroy()

            /* When */
            navigator1.push(navigator1Scene2)

            /* Then */
            verify(listener, never()).scene(eq(navigator1Scene2), anyOrNull())
        }

        @Test
        fun `a stopped navigator propagates child navigator emitting scene when started`() {
            /* Given */
            navigator.addNavigatorEventsListener(listener)
            navigator.onStart()
            navigator.onStop()

            /* When */
            navigator1.push(navigator1Scene2)
            navigator.onStart()

            /* Then */
            verify(listener).scene(eq(navigator1Scene2), anyOrNull())
        }

        @Test
        fun `a stopped navigator propagates child navigator emitting scene when started for late listeners`() {
            /* Given */
            navigator.onStart()
            navigator.addNavigatorEventsListener(listener)
            navigator.onStop()

            /* When */
            navigator1.push(navigator1Scene2)
            navigator.onStart()

            /* Then */
            verify(listener).scene(eq(navigator1Scene2), anyOrNull())
        }

        @Test
        fun `a removed listener during active state does not get notified of scenes anymore when stopped and started`() {
            /* Given */
            val disposable = navigator.addNavigatorEventsListener(listener)
            navigator.onStart()
            disposable.dispose()
            navigator.onStop()

            /* When */
            navigator1.push(navigator1Scene2)
            navigator.onStart()

            /* Then */
            verify(listener, never()).scene(eq(navigator1Scene2), anyOrNull())
        }
    }

    @Nested
    inner class MultipleChildrenNavigatorInteraction {

        @Test
        fun `changing destination for inactive navigator does not start selected child navigator`() {
            /* When */
            navigator.select(Destination2)

            /* Then */
            verify(navigator2, never()).onStart()
        }

        @Test
        fun `starting navigator after destination change starts selected child navigator`() {
            /* Given */
            navigator.select(Destination2)

            /* When */
            navigator.onStart()

            /* Then */
            verify(navigator2).onStart()
        }

        @Test
        fun `changing destination after navigator started starts selected child navigator`() {
            /* Given */
            navigator.onStart()

            /* When */
            navigator.select(Destination2)

            /* Then */
            verify(navigator2).onStart()
        }

        @Test
        fun `changing to the same destination twice after navigator started starts selected child navigator only once`() {
            /* Given */
            navigator.onStart()

            /* When */
            navigator.select(Destination2)
            navigator.select(Destination2)

            /* Then */
            verify(navigator2, times(1)).onStart()
        }

        @Test
        fun `starting navigator after destination change does not start initial child navigator`() {
            /* Given */
            navigator.select(Destination2)

            /* When */
            navigator.onStart()

            /* Then */
            verify(navigator1, never()).onStart()
        }

        @Test
        fun `changing destination after navigator started stops initial child navigator`() {
            /* Given */
            navigator.onStart()

            /* When */
            navigator.select(Destination2)

            /* Then */
            navigator1.inOrder {
                verify().onStart()
                verify().onStop()
                verifyNoMoreInteractions()
            }
        }

        @Test
        fun `stopping inactive navigator does not stop children`() {
            /* Given */
            navigator.select(Destination2)

            /* When */
            navigator.onStop()

            /* Then */
            verify(navigator1, never()).onStop()
            verify(navigator2, never()).onStop()
        }

        @Test
        fun `stopping active navigator stops only selected navigator`() {
            /* Given */
            navigator.select(Destination2)
            navigator.onStart()

            /* When */
            navigator.onStop()

            /* Then */
            verify(navigator1, never()).onStop()
            navigator2.inOrder {
                verify().onStart()
                verify().onStop()
            }
        }

        @Test
        fun `stopping destroyed navigator does not stop children`() {
            /* Given */
            navigator.select(Destination2)
            navigator.onDestroy()

            /* When */
            navigator.onStop()

            /* Then */
            verify(navigator1, never()).onStop()
            verify(navigator2, never()).onStop()
        }

        @Test
        fun `destroying inactive navigator destroys all children but does not stop them`() {
            /* Given */
            navigator.select(Destination2)

            /* When */
            navigator.onDestroy()

            /* Then */
            verify(navigator1, never()).onStop()
            verify(navigator1).onDestroy()
            verify(navigator2, never()).onStop()
            verify(navigator2).onDestroy()
        }

        @Test
        fun `destroying active navigator stops selected child and destroys all children`() {
            /* Given */
            navigator.select(Destination2)
            navigator.onStart()

            /* When */
            navigator.onDestroy()

            /* Then */
            verify(navigator1, never()).onStop()
            verify(navigator1).onDestroy()
            verify(navigator2).onStop()
            verify(navigator2).onDestroy()
        }

        @Test
        fun `destroying destroyed navigator does not destroy child navigators again`() {
            /* Given */
            navigator.select(Destination2)
            navigator.onDestroy()

            /* When */
            navigator.onDestroy()

            /* Then */
            verify(navigator1, times(1)).onDestroy()
            verify(navigator2, times(1)).onDestroy()
        }

        @Test
        fun `changing destination propagates scene for selected destination`() {
            /* Given */
            navigator.addNavigatorEventsListener(listener)
            navigator.onStart()

            /* When */
            navigator.select(Destination2)

            /* Then */
            listener.inOrder {
                verify().scene(eq(navigator1Scene1), anyOrNull())
                verify().scene(eq(navigator2Scene1), anyOrNull())
            }
        }

        @Test
        fun `changing destination multiple times propagates scene for selected destinations`() {
            /* Given */
            navigator.addNavigatorEventsListener(listener)
            navigator.onStart()

            /* When */
            navigator.select(Destination2)
            navigator.select(Destination1)

            /* Then */
            listener.inOrder {
                verify().scene(eq(navigator1Scene1), anyOrNull())
                verify().scene(eq(navigator2Scene1), anyOrNull())
                verify().scene(eq(navigator1Scene1), anyOrNull())
            }
        }

        @Test
        fun `an active navigator propagates selected child navigator emitting scene`() {
            /* Given */
            navigator.addNavigatorEventsListener(listener)
            navigator.select(Destination2)
            navigator.onStart()

            /* When */
            navigator2.push(navigator2Scene2)

            /* Then */
            verify(listener).scene(eq(navigator2Scene2), anyOrNull())
        }

        @Test
        fun `an active navigator does not propagate non-selected child navigator emitting scene`() {
            /* Given */
            navigator.addNavigatorEventsListener(listener)
            navigator.select(Destination2)
            navigator.onStart()

            /* When */
            navigator1.push(navigator1Scene2)

            /* Then */
            verify(listener).scene(eq(navigator2Scene1), anyOrNull())
            verify(listener, never()).scene(eq(navigator1Scene2), anyOrNull())
        }
    }

    @Nested
    inner class ChildInteractionOnBackPress {

        @Test
        fun `onBackPressed for a single scene in first destination for inactive navigator destroys child navigator`() {
            /* When */
            navigator.onBackPressed()

            /* Then */
            verify(navigator1).onDestroy()
        }

        @Test
        fun `onBackPressed for a single scene in first destination for inactive navigator destroys parent navigator`() {
            /* When */
            navigator.onBackPressed()

            /* When */
            expect(navigator.isDestroyed()).toBe(true)
        }

        @Test
        fun `onBackPressed for a single scene in first destination for inactive navigator does not stop child navigator`() {
            /* When */
            navigator.onBackPressed()

            /* When */
            verify(navigator1, never()).onStop()
        }

        @Test
        fun `onBackPressed for a single scene in second destination for inactive navigator destroys second child navigator`() {
            /* Given */
            navigator.select(Destination2)

            /* When */
            navigator.onBackPressed()

            /* When */
            verify(navigator2).onDestroy()
        }

        @Test
        fun `onBackPressed for a single scene in first destination for active navigator stops and destroys child navigator`() {
            /* Given */
            navigator.onStart()

            /* When */
            navigator.onBackPressed()

            /* When */
            navigator1.inOrder {
                verify().onStop()
                verify().onDestroy()
            }
        }

        @Test
        fun `onBackPressed for a single scene in first destination for active navigator destroys parent navigator`() {
            /* Given */
            navigator.onStart()

            /* When */
            navigator.onBackPressed()

            /* When */
            expect(navigator.isDestroyed()).toBe(true)
        }

        @Test
        fun `onBackPressed for a single scene in second destination for active navigator stops and destroys second child navigator, and starts current child navigator`() {
            /* Given */
            navigator.onStart()
            navigator.select(Destination2)

            /* When */
            navigator.onBackPressed()

            /* When */
            inOrder(navigator1, navigator2) {
                verify(navigator2).onStop()
                verify(navigator2).onDestroy()
                verify(navigator1).onStart()
            }
        }

        @Test
        fun `onBackPressed for a single scene in second destination active navigator does not stop or destroy second child navigator`() {
            /* Given */
            navigator.onStart()
            navigator.select(Destination2)

            /* When */
            navigator2.push(navigator2Scene2)
            navigator.onBackPressed()

            /* When */
            verify(navigator2, never()).onStop()
        }

        @Test
        fun `onBackPressed for a primary child navigator that does not support back press during inactive state`() {
            /* Given */
            val childNavigator = NonBackPressableNavigator()
            val navigator = SavableTestCompositeParallelNavigator(mapOf(Destination1 to childNavigator))
            navigator.addNavigatorEventsListener(listener)

            /* When */
            val result = navigator.onBackPressed()

            /* Then */
            expect(result).toBe(true)
            verify(listener).finished()
        }

        @Test
        fun `onBackPressed for non-primary child navigator that does not support back press during inactive state`() {
            /* Given */
            val childNavigator = NonBackPressableNavigator()
            val navigator = SavableTestCompositeParallelNavigator(
                mapOf(
                    Destination1 to navigator1,
                    Destination2 to childNavigator
                )
            )
            navigator.select(Destination2)
            navigator.addNavigatorEventsListener(listener)

            /* When */
            val result = navigator.onBackPressed()

            /* Then */
            expect(result).toBe(true)
            verify(listener, never()).finished()

            /* When */
            navigator.onStart()

            /* Then */
            verify(listener).scene(eq(navigator1Scene1), anyOrNull())
        }

        @Test
        fun `onBackPressed for a primary child navigator that does not support back press during active state`() {
            /* Given */
            val childNavigator = NonBackPressableNavigator()
            val navigator = SavableTestCompositeParallelNavigator(mapOf(Destination1 to childNavigator))
            navigator.onStart()
            navigator.addNavigatorEventsListener(listener)

            /* When */
            val result = navigator.onBackPressed()

            /* Then */
            expect(result).toBe(true)
            verify(listener).finished()
        }

        @Test
        fun `onBackPressed for a destroyed navigator returns false`() {
            /* Given */
            navigator.onDestroy()

            /* When */
            val result = navigator.onBackPressed()

            /* Then */
            expect(result).toBe(false)
        }
    }

    @Nested
    inner class ChildrenFinishing {

        @Nested
        inner class Inactive {

            @Test
            fun `the primary destination finishing while selected finishes parent navigator`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)

                /* When */
                navigator1.finish()

                /* Then */
                verify(listener).finished()
            }

            @Test
            fun `the primary destination finishing while selected destroys child navigator for parent navigator`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)

                /* When */
                navigator1.finish()

                /* Then */
                verify(navigator1).onDestroy()
            }

            @Test
            fun `the primary destination finishing while not selected does not finish parent navigator`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.select(Destination2)

                /* When */
                navigator1.finish()

                /* Then */
                verify(listener, never()).finished()
            }

            @Test
            fun `a non-primary, non-selected destination finishing does not finish parent navigator`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.select(Destination2)
                navigator.select(Destination1)

                /* When */
                navigator2.finish()

                /* Then */
                verify(listener, never()).finished()
            }

            @Test
            fun `a non-primary, non-selected destination finishing destroys child navigator for parent navigator`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.select(Destination2)
                navigator.select(Destination1)

                /* When */
                navigator2.finish()

                /* Then */
                verify(navigator2).onDestroy()
            }

            @Test
            fun `a non-primary, selected destination finishing selects initial navigator when started`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.select(Destination2)

                /* When */
                navigator2.finish()

                /* Then */
                verify(listener, never()).scene(eq(navigator1Scene1), anyOrNull())

                /* When */
                navigator.onStart()

                /* Then */
                listener.inOrder {
                    verify().scene(eq(navigator1Scene1), anyOrNull())
                    verifyNoMoreInteractions()
                }
            }

            @Test
            fun `selecting a finished primary navigator recreates it when started`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.select(Destination2)

                /* When */
                navigator1.finish()
                navigator1 = spy(SavableTestStackNavigator(listOf(extraNavigator)))
                navigator.select(Destination1)

                /* Then */
                verify(listener, never()).scene(eq(extraNavigator), anyOrNull())

                /* When */
                navigator.onStart()

                /* Then */
                listener.inOrder {
                    verify().scene(eq(extraNavigator), anyOrNull())
                    verifyNoMoreInteractions()
                }
            }

            @Test
            fun `selecting a finished non-primary navigator recreates it when started`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.select(Destination2)
                navigator.select(Destination1)

                /* When */
                navigator2.finish()
                navigator2 = spy(SavableTestStackNavigator(listOf(extraNavigator)))
                navigator.select(Destination2)

                /* Then */
                verify(listener, never()).scene(eq(extraNavigator), anyOrNull())

                /* When */
                navigator.onStart()

                /* Then */
                listener.inOrder {
                    verify().scene(eq(extraNavigator), anyOrNull())
                    verifyNoMoreInteractions()
                }
            }
        }

        @Nested
        inner class Active {

            @Test
            fun `the primary destination finishing while selected finishes parent navigator`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.onStart()

                /* When */
                navigator1.finish()

                /* Then */
                verify(listener).finished()
            }

            @Test
            fun `the primary destination finishing while selected stops and destroys child navigator for parent navigator`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.onStart()

                /* When */
                navigator1.finish()

                /* Then */
                navigator1.inOrder {
                    verify().onStop()
                    verify().onDestroy()
                }
            }

            @Test
            fun `the primary destination finishing while not selected does not finish parent navigator`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.onStart()
                navigator.select(Destination2)

                /* When */
                navigator1.finish()

                /* Then */
                verify(listener, never()).finished()
            }

            @Test
            fun `a non-primary, non-selected destination finishing does not finish parent navigator`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.select(Destination2)
                navigator.select(Destination1)
                navigator.onStart()

                /* When */
                navigator2.finish()

                /* Then */
                verify(listener, never()).finished()
            }

            @Test
            fun `a non-primary, non-selected destination finishing stops and destroys child navigator for parent navigator`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.select(Destination2)
                navigator.select(Destination1)
                navigator.onStart()

                /* When */
                navigator2.finish()

                /* Then */
                navigator2.inOrder {
                    verify().onStop()
                    verify().onDestroy()
                }
            }

            @Test
            fun `a non-primary, selected destination finishing selects initial navigator`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.select(Destination2)
                navigator.onStart()

                /* When */
                navigator2.finish()

                /* Then */
                listener.inOrder {
                    verify().scene(eq(navigator2Scene1), anyOrNull())
                    verify().scene(eq(navigator1Scene1), anyOrNull())
                }
            }

            @Test
            fun `a non-primary, non-selected destination finishing does not select initial navigator`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.select(Destination3)
                navigator.select(Destination2)
                navigator.onStart()

                /* When */
                navigator3.finish()

                /* Then */
                listener.inOrder {
                    verify().scene(eq(navigator2Scene1), anyOrNull())
                    verifyNoMoreInteractions()
                }
            }

            @Test
            fun `a non-primary, selected destination finishing starts initial navigator`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.select(Destination2)
                navigator.onStart()

                /* When */
                navigator2.finish()

                /* Then */
                inOrder(navigator1, navigator2) {
                    verify(navigator2).onStop()
                    verify(navigator2).onDestroy()
                    verify(navigator1).onStart()
                }
            }

            @Test
            fun `selecting a finished primary navigator recreates it`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.select(Destination2)
                navigator.onStart()

                /* When */
                navigator1.finish()
                navigator1 = spy(SavableTestStackNavigator(listOf(extraNavigator)))
                navigator.select(Destination1)

                /* Then */
                listener.inOrder {
                    verify().scene(eq(extraNavigator), anyOrNull())
                    verifyNoMoreInteractions()
                }
            }

            @Test
            fun `selecting a finished non-primary navigator recreates it`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.select(Destination2)
                navigator.select(Destination1)
                navigator.onStart()

                /* When */
                navigator2.finish()
                navigator2 = spy(SavableTestStackNavigator(listOf(extraNavigator)))
                navigator.select(Destination2)

                /* Then */
                listener.inOrder {
                    verify().scene(eq(extraNavigator), anyOrNull())
                    verifyNoMoreInteractions()
                }
            }
        }

        @Nested
        inner class Destroyed {

            @BeforeEach
            fun beforeEach() {
                navigator.onDestroy()
            }

            @Test
            fun `the primary destination finishing does not notify listener`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)

                /* When */
                navigator1.finish()

                /* Then */
                verifyNoMoreInteractions(listener)
            }

            @Test
            fun `the primary destination finishing does not stop or destroy child navigator`() {
                /* Given */
                inOrder(navigator1) {
                    verify(navigator1).onDestroy()
                    navigator.addNavigatorEventsListener(listener)

                    /* When */
                    navigator1.finish()

                    /* Then */
                    verify(navigator1, never()).onStop()
                    verify(navigator1, never()).onDestroy()
                }
            }

            @Test
            fun `the primary destination finishing while not selected does not finish parent navigator`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.select(Destination2)

                /* When */
                navigator1.finish()

                /* Then */
                verify(listener, never()).finished()
            }

            @Test
            fun `a non-primary, non-selected destination finishing does not finish parent navigator`() {
                /* Given */
                navigator.addNavigatorEventsListener(listener)
                navigator.select(Destination2)
                navigator.select(Destination1)

                /* When */
                navigator2.finish()

                /* Then */
                verify(listener, never()).finished()
            }
        }
    }

    @Nested
    inner class SavingState {

        private val navigator1Scene1 = SavableTestScene(11)
        private val navigator2Scene1 = SavableTestScene(21)

        private val navigator1 = TestSingleSceneNavigator(navigator1Scene1)
        private val navigator2 = TestSingleSceneNavigator(navigator2Scene1)
        private val savableNavigator1 = SavableTestSingleSceneNavigator(navigator1Scene1)
        private val savableNavigator2 = SavableTestSingleSceneNavigator(navigator2Scene1)

        @Test
        fun `CompositeParallelNavigator does not implement SavableNavigator by default`() {
            /* Given */
            val navigator: Navigator = TestCompositeParallelNavigator()

            /* Then */
            expect(navigator is SavableNavigator).toBe(false)
        }

        @Test
        fun `saving and restoring state for single savable navigator`() {
            /* Given */
            val navigator = SavableTestCompositeParallelNavigator(
                mapOf(Destination1 to savableNavigator1),
                null
            )
            navigator.onStart()
            navigator1Scene1.foo = 3

            /* When */
            val bundle = navigator.saveInstanceState()
            navigator1Scene1.foo = 42

            val restoredNavigator = SavableTestCompositeParallelNavigator(
                mapOf(Destination1 to savableNavigator1),
                bundle
            )
            restoredNavigator.onStart()
            restoredNavigator.addNavigatorEventsListener(listener)

            /* Then */
            argumentCaptor<Scene<out Container>> {
                verify(listener).scene(capture(), anyOrNull())
                expect(lastValue).toNotBeTheSameAs(navigator1Scene1)
                expect(lastValue).toBeInstanceOf<SavableTestScene> {
                    expect(it.foo).toBe(3)
                }
            }
        }

        @Test
        fun `saving and restoring state for single non savable navigator`() {
            /* Given */
            val navigator = SavableTestCompositeParallelNavigator(
                mapOf(Destination1 to navigator1),
                null
            )
            navigator.onStart()
            navigator1Scene1.foo = 3

            /* When */
            val bundle = navigator.saveInstanceState()
            navigator1Scene1.foo = 42

            val restoredNavigator = SavableTestCompositeParallelNavigator(
                mapOf(Destination1 to navigator2),
                bundle
            )
            restoredNavigator.onStart()
            restoredNavigator.addNavigatorEventsListener(listener)

            /* Then */
            argumentCaptor<Scene<out Container>> {
                verify(listener).scene(capture(), anyOrNull())
                expect(lastValue).toBeInstanceOf<SavableTestScene> {
                    expect(it.foo).toBe(21)
                }
            }
        }

        @Test
        fun `saving and restoring state for multiple savable navigators`() {
            /* Given */
            val navigator = SavableTestCompositeParallelNavigator(
                mapOf(
                    Destination1 to savableNavigator1,
                    Destination2 to savableNavigator2
                ),
                null
            )
            navigator.onStart()
            navigator1Scene1.foo = 3
            navigator.select(Destination2)
            navigator2Scene1.foo = 4

            /* When */
            val bundle = navigator.saveInstanceState()
            navigator2Scene1.foo = 42

            val restoredNavigator = SavableTestCompositeParallelNavigator(
                mapOf(
                    Destination1 to savableNavigator1,
                    Destination2 to savableNavigator2
                ),
                bundle
            )
            restoredNavigator.onStart()
            restoredNavigator.addNavigatorEventsListener(listener)

            /* Then */
            argumentCaptor<Scene<out Container>> {
                verify(listener).scene(capture(), anyOrNull())
                expect(lastValue).toNotBeTheSameAs(navigator2Scene1)
                expect(lastValue).toBeInstanceOf<SavableTestScene> {
                    expect(it.foo).toBe(4)
                }
            }
        }

        @Test
        fun `restoring from malformed state falls back to initial state - empty state`() {
            /* Given */
            val bundle = NavigatorState()

            /* When */
            val restoredNavigator = SavableTestCompositeParallelNavigator(
                mapOf(
                    Destination1 to savableNavigator1,
                    Destination2 to savableNavigator2
                ),
                bundle
            )
            restoredNavigator.onStart()
            restoredNavigator.addNavigatorEventsListener(listener)

            /* Then */
            argumentCaptor<Scene<out Container>> {
                verify(listener).scene(capture(), anyOrNull())
                expect(lastValue).toBeTheSameAs(navigator1Scene1)
                expect(lastValue).toBeInstanceOf<SavableTestScene> {
                    expect(it.foo).toBe(11)
                }
            }
        }

        @Test
        fun `restoring from malformed state falls back to initial state - wrong size`() {
            /* Given */
            val bundle = navigatorState {
                it["size"] = 2
            }

            /* When */
            val restoredNavigator = SavableTestCompositeParallelNavigator(
                mapOf(
                    Destination1 to savableNavigator1,
                    Destination2 to savableNavigator2
                ),
                bundle
            )
            restoredNavigator.onStart()
            restoredNavigator.addNavigatorEventsListener(listener)

            /* Then */
            argumentCaptor<Scene<out Container>> {
                verify(listener).scene(capture(), anyOrNull())
                expect(lastValue).toBeTheSameAs(navigator1Scene1)
                expect(lastValue).toBeInstanceOf<SavableTestScene> {
                    expect(it.foo).toBe(11)
                }
            }
        }

        @Test
        fun `restoring from malformed state falls back to initial state - missing destination`() {
            /* Given */
            val bundle = navigatorState {
                it["size"] = 1
                it["0_state"] = NavigatorState()
                it["0_destination"] = "0"
            }

            /* When */
            val restoredNavigator = SavableTestCompositeParallelNavigator(
                mapOf(
                    Destination1 to savableNavigator1,
                    Destination2 to savableNavigator2
                ),
                bundle
            )
            restoredNavigator.onStart()
            restoredNavigator.addNavigatorEventsListener(listener)

            /* Then */
            argumentCaptor<Scene<out Container>> {
                verify(listener).scene(capture(), anyOrNull())
                expect(lastValue).toBeTheSameAs(navigator1Scene1)
                expect(lastValue).toBeInstanceOf<SavableTestScene> {
                    expect(it.foo).toBe(11)
                }
            }
        }

        @Test
        fun `saved state from callback is the same as saved state _after_ callback`() {
            /* Given */
            val navigator = SavableTestCompositeParallelNavigator(
                mapOf(
                    Destination1 to savableNavigator1,
                    Destination2 to savableNavigator2
                ),
                null
            )

            var state1: SavedState? = null
            navigator.addNavigatorEventsListener(object : Navigator.Events {
                override fun scene(scene: Scene<out Container>, data: TransitionData?) {
                    state1 = navigator.saveInstanceState()
                }

                override fun finished() {
                }
            })
            navigator.onStart()

            /* When */
            navigator.select(Destination2)
            val state2 = navigator.saveInstanceState()

            /* Then */
            expect(state1).toBe(state2)
        }

        @Test
        fun `saving a destroyed navigator results in empty state`() {
            /* Given */
            val navigator = SavableTestCompositeParallelNavigator(
                mapOf(Destination1 to savableNavigator1),
                null
            )
            navigator.onStart()
            navigator.onStop()
            navigator.onDestroy()

            /* When */
            val state = navigator.saveInstanceState()

            /* Then */
            expect(state.entries.size).toBe(0)
        }
    }

    private enum class TestDestination {
        Destination1,
        Destination2,
        Destination3
    }

    private inner class TestCompositeParallelNavigator : CompositeParallelNavigator<TestDestination>(Destination1, null) {

        override fun serialize(destination: TestDestination): String {
            error("Not used")
        }

        override fun deserialize(serializedDestination: String): TestDestination {
            error("Not used")
        }

        override fun createNavigator(destination: TestDestination, savedState: NavigatorState?): Navigator {
            return when (destination) {
                Destination1 -> navigator1
                Destination2 -> navigator2
                Destination3 -> navigator3
            }
        }
    }

    private inner class SavableTestCompositeParallelNavigator(
        private val navigators: Map<TestDestination, Navigator>,
        savedState: NavigatorState? = null
    ) : CompositeParallelNavigator<TestDestination>(Destination1, savedState), SavableNavigator {

        override fun serialize(destination: TestDestination): String {
            return "${destination.ordinal}"
        }

        override fun deserialize(serializedDestination: String): TestDestination {
            return values()[serializedDestination.toInt()]
        }

        override fun createNavigator(destination: TestDestination, savedState: NavigatorState?): Navigator {
            if (savedState != null) {
                return when (destination) {
                    Destination1 -> SavableTestSingleSceneNavigator(navigator1Scene1, savedState)
                    Destination2 -> SavableTestSingleSceneNavigator(navigator2Scene1, savedState)
                    Destination3 -> SavableTestSingleSceneNavigator(navigator3Scene1, savedState)
                }
            }

            return navigators[destination]!!
        }
    }

    private class NonBackPressableNavigator : Navigator {

        override fun addNavigatorEventsListener(listener: Navigator.Events): DisposableHandle {
            return object : DisposableHandle {
                private var disposed = false
                override fun dispose() {
                    disposed = true
                }

                override fun isDisposed(): Boolean {
                    return disposed
                }
            }
        }

        override fun onStart() {
        }

        override fun onStop() {
        }

        private var isDestroyed = false
        override fun onDestroy() {
            isDestroyed = true
        }

        override fun isDestroyed(): Boolean {
            return isDestroyed
        }
    }

    private open class SavableTestStackNavigator(
        private val initialStack: List<SavableTestScene>,
        savedState: NavigatorState? = null
    ) : StackNavigator(savedState), SavableNavigator {

        override fun initialStack(): List<Scene<out Container>> {
            return initialStack
        }

        override fun instantiateScene(sceneClass: KClass<out Scene<*>>, state: SceneState?): Scene<*> {
            return when (sceneClass) {
                SavableTestScene::class -> SavableTestScene.create(state)
                else -> error("Unknown class: $sceneClass")
            }
        }
    }

    open class TestSingleSceneNavigator(
        private val scene: Scene<out Container>
    ) : SingleSceneNavigator(null) {

        override fun createScene(state: SceneState?): Scene<out Container> {
            return scene
        }
    }

    open class SavableTestSingleSceneNavigator(
        private val scene: Scene<out Container>,
        savedState: NavigatorState? = null
    ) : SingleSceneNavigator(savedState), SavableNavigator {

        override fun createScene(state: SceneState?): Scene<out Container> {
            return state?.let { SavableTestScene.create(it) } ?: scene
        }
    }
}
